#ifndef CFFREAD_H
#define CFFREAD_H

typedef struct cff_compoundfile_struct COMPOUNDFILE;

/* mode는 fopen에 그대로 전달된다.  */
COMPOUNDFILE *cff_open(const char *path, const char *mode);
void cff_close(COMPOUNDFILE *cfp);

typedef struct {
    size_t alloc_size; /* data에 할당된 크기 */
    size_t data_size;  /* data의 내용물 크기 */
    void *data;
} CFFSTREAM;

/* cff_salloc: 복합파일 안에서 스트림의 경로를 받아서 스트림
   포인터를 반환한다.  경로의 형식은 유닉스와 비슷하다.
   최상위 스토리지는 루트이고 모든 스토리지와 스트림은 루트에 속한다.
   HWP5.0 파일의 본문이 저장된 Section0 스트림의 예를 들면
        /BodyText/Section0
   가 경로 문자열이 된다.  점경로를 제외한 상대경로도 지원하므로
        BodyText/Section0
   로 쓸 수도 있다.  다쓴 스트림은 cff_free로 정리한다.  */
CFFSTREAM *cff_salloc(const char *stream_path, COMPOUNDFILE *cfp);
/* directory id로 스트림을 할당한다.  cff_free와 같이 쓴다.  */
CFFSTREAM *cff_salloc_by_id(unsigned int id, COMPOUNDFILE *cfp);
void cff_free(CFFSTREAM *stream);

/* cff_change_storage: 복합파일 안에서 해당 스토리지로 들어간다.
   리눅스 cd 명령어와 비슷하다.  점경로는 사용하지 않는다.
   상대경로 또는 절대경로로 스토리지를 지정할 수 있다.
   처음에 복합파일을 열면 루트 스토리지가 현재 스토리지이다.
   필요할 것 같아서 만들었지만 한번도 사용하지 않았다.  */
int cff_change_storage(const char *path, COMPOUNDFILE *cfp);
unsigned int cff_get_dirid(const char *path, COMPOUNDFILE *cfp);

#endif /* CFFREAD_H */
