#ifndef CFFUTIL_H
#define CFFUTIL_H

#include "cffread.h"
#include "linkedlist.h"

/* LNode 리스트의 data에 붙이는 구조체이다.  */
typedef struct _CffDirLNode CffDirLNode;
struct _CffDirLNode {
    unsigned int id;        /* directory id */
    unsigned int parent_id; /* parent directory id */
    char *name;             /* stream or storage name */
    char *path;             /* stream or storage path */
    unsigned char type;     /* type: root, stream, storage */
};

/* 스트림과 스토리지 목록을 반환한다.
   경로와 이름은 utf-8 또는 c locale로 저장했다.
   반환된 포인터는 cff_destroy_dir_list로 정리한다.  */
LNode *cff_make_dir_list(COMPOUNDFILE *cfp);
void cff_destroy_dir_list(LNode *base);
CffDirLNode *cff_find_dir_node(const char *name, LNode *base);
CffDirLNode *cff_find_dir_node_by_id(unsigned int id, LNode *base);

typedef CFFSTREAM *(*CffStreamFilter)(const CFFSTREAM *);
/* 복합 파일을 일반 파일로 나눈다.  스트림을 파일로 저장하기
   전에 하고 싶은 작업은 CffStreamFilter 형식의 함수 포인터로
   지정할 수 있다.  필요 없다면 널을 넣는다.
   node: 리스트에서 스트림이나 스토리지의 노드
   base: cff_make_dir_list에서 받은 포인터
   cfp: 복합파일
   filter: 스트림을 처리하는 함수 포인터
   no_overwrite: 덮어쓰려면 0, 덮어쓰지 않으려면 0이 아닌 값
   path: 접두어로 쓸 복합 파일 이름, 쓰지 않으려면 빈문자열. */
void cff_make_split_stream_storage(CffDirLNode *node,
                                   LNode *base,
                                   COMPOUNDFILE *cfp,
                                   CffStreamFilter filter,
                                   int no_overwrite,
                                   const char *path);

#endif /* CFFUTIL_H */
