#ifndef FILEUTIL_H
#define FILEUTIL_H

/* path와 suffix를 받아서 파일 이름을 만든다. 
   path의 파일명과 확장자와 suffix를 연결시킨다.
   파일명과 확장자 사이의 점은 밑줄로 바꾼다.
   path가 빈문자열이 아니면 suffix 앞에 밑줄이 붙는다.
   반환문자열은 쓰고 해제한다.  */
char *make_filename(const char *path, const char *suffix);

/* path 디렉토리로 이동한다.  없으면 만들고 이동한다.
   같은 이름의 다른 파일이 있으면 실패한다.  반환: 성공 0, 실패 -1
   실패시 메시지 출력  */
int make_dir_chdir(const char *path);
/* make_filename로 만든 이름을 이용한다.  */
int make_dir_chdir_with_suffix(const char *path, const char *suffix);

/* 슬래시나 역슬래시로 디렉토리를 제외한 파일명을 반환한다.
   반환문자열은 쓰고 해제한다.  */
char *strdup_basename(const char *path);

/* make_dump_filename을 이용해서 파일 이름을 정하고 거기로
   표준출력을 돌린다.  성공 0, 실패 -1
   에러시 메시지 출력.  */
int redirect_stdout_to_file(int no_overwrite,
                            const char *path,
                            const char *suffix);

/* no_overwrite가 0일 때 파일이 이미 있어도 0을 반환한다.
   no_overwrite가 0이 아닐 때 파일이 이미 있으면 -1을 반환한다.
   에러가 발생하면 -1을 반환한다.
   -1을 반환할 때는 적절한 메시지를 같이 출력한다.  */
int check_file(const char *path, int no_overwrite);

/* 파일, 디렉토리와 관련된 errno 메시지를 출력한다.  */
void file_err_msg(int err, const char *path);

char *strdup_pwd(void);

#endif /* FILEUTIL_H */
