/* hwp5dump.c

   hwp5dump는 한글과컴퓨터의 한글 문서 파일 구조 정의
   공개 정책을 따르는 2차 저작물입니다.  */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "cffread.h"
#include "cffutil.h"
#include "hwp5read.h"
#include "hwp5type.h"
#include "dumputil.h"
#include "fileutil.h"
#include "unicode_locale.h"
#include "unicode_string.h"

#define PROGRAM_NAME_AND_VERSION "hwp5dump 0.1"

#define IS_OPTION(x) ((option & (x)) == (x))
static int option = 0;
enum _hwp5dump_option {
    OVERWRITE = 1,
    DESTINATION_DIRECTORY = 1 << 1,
    STANDARD_OUTPUT = 1 << 2,
    TEXT_WITHOUT_HEX = 1 << 3,
    BINDATA_FILE = 1 << 4
};
static const char *destination_directory = "";

void  hwp5_print_header  (HWP5FILE *hfp);
void  hwp5_print_header_attribute(DWORD attr);

void  hwp5_print_summary (HWP5FILE *hfp);
void  hwp5_print_prvtext (HWP5FILE *hfp);

void hwp5_print_hdrstream(const char *path, HWP5FILE *hfp);
void hwp5_print_hdr_list(const char *path, int size, HDRSTREAM *h);
void hwp5_print_hdr(int n, HDR *hdr);
void hwp5_print_hdr_table(const char *path, int size, HDRSTREAM *h);
void hwp5_print_hdr_row(int n, HDR *hdr);

void hwp5_print_body_text(HWP5FILE *hfp);
void hwp5_print_section_text(HDRSTREAM *s);
void hwp5_print_para_text(HDR *p);

void hwp5_extract_bindata_file(const char *path, HWP5FILE *hfp);


char *hdr_tag_name(int tag_id)
{
    switch (tag_id) {
    /* 문서 정보 스트림의 태그 */
    case HWPTAG_DOCUMENT_PROPERTIES:
        return "HWPTAG_DOCUMENT_PROPERTIES";
    case HWPTAG_ID_MAPPINGS:
        return "HWPTAG_ID_MAPPINGS";
    case HWPTAG_BIN_DATA:
        return "HWPTAG_BIN_DATA";
    case HWPTAG_FACE_NAME:
        return "HWPTAG_FACE_NAME";
    case HWPTAG_BORDER_FILL:
        return "HWPTAG_BORDER_FILL";
    case HWPTAG_CHAR_SHAPE:
        return "HWPTAG_CHAR_SHAPE";
    case HWPTAG_TAB_DEF:
        return "HWPTAG_TAB_DEF";
    case HWPTAG_NUMBERING:
        return "HWPTAG_NUMBERING";
    case HWPTAG_BULLET:
        return "HWPTAG_BULLET";
    case HWPTAG_PARA_SHAPE:
        return "HWPTAG_PARA_SHAPE";
    case HWPTAG_STYLE:
        return "HWPTAG_STYLE";
    case HWPTAG_DOC_DATA:
        return "HWPTAG_DOC_DATA";
    case HWPTAG_DISTRIBUTE_DOC_DATA:
        return "HWPTAG_DISTRIBUTE_DOC_DATA";
    case HWPTAG_COMPATIBLE_DOCUMENT:
        return "HWPTAG_COMPATIBLE_DOCUMENT";
    case HWPTAG_LAYOUT_COMPATIBILITY:
        return "HWPTAG_LAYOUT_COMPATIBILITY";
    case HWPTAG_FORBIDDEN_CHAR:
        return "HWPTAG_FORBIDDEN_CHAR";

    /* 본문 태그 */
    case HWPTAG_PARA_HEADER:
        return "HWPTAG_PARA_HEADER";
    case HWPTAG_PARA_TEXT:
        return "HWPTAG_PARA_TEXT";
    case HWPTAG_PARA_CHAR_SHAPE:
        return "HWPTAG_PARA_CHAR_SHAPE";
    case HWPTAG_PARA_LINE_SEG:
        return "HWPTAG_PARA_LINE_SEG";
    case HWPTAG_PARA_RANGE_TAG:
        return "HWPTAG_PARA_RANGE_TAG";
    case HWPTAG_CTRL_HEADER:
        return "HWPTAG_CTRL_HEADER";
    case HWPTAG_LIST_HEADER:
        return "HWPTAG_LIST_HEADER";
    case HWPTAG_PAGE_DEF:
        return "HWPTAG_PAGE_DEF";
    case HWPTAG_FOOTNOTE_SHAPE:
        return "HWPTAG_FOOTNOTE_SHAPE";
    case HWPTAG_PAGE_BORDER_FILL:
        return "HWPTAG_PAGE_BORDER_FILL";
    case HWPTAG_SHAPE_COMPONENT:
        return "HWPTAG_SHAPE_COMPONENT";
    case HWPTAG_TABLE:
        return "HWPTAG_TABLE";
    case HWPTAG_SHAPE_COMPONENT_LINE:
        return "HWPTAG_SHAPE_COMPONENT_LINE";
    case HWPTAG_SHAPE_COMPONENT_RECTANGLE:
        return "HWPTAG_SHAPE_COMPONENT_RECTANGLE";
    case HWPTAG_SHAPE_COMPONENT_ELLIPSE:
        return "HWPTAG_SHAPE_COMPONENT_ELLIPSE";
    case HWPTAG_SHAPE_COMPONENT_ARC:
        return "HWPTAG_SHAPE_COMPONENT_ARC";
    case HWPTAG_SHAPE_COMPONENT_POLYGON:
        return "HWPTAG_SHAPE_COMPONENT_POLYGON";
    case HWPTAG_SHAPE_COMPONENT_CURVE:
        return "HWPTAG_SHAPE_COMPONENT_CURVE";
    case HWPTAG_SHAPE_COMPONENT_OLE:
        return "HWPTAG_SHAPE_COMPONENT_OLE";
    case HWPTAG_SHAPE_COMPONENT_PICTURE:
        return "HWPTAG_SHAPE_COMPONENT_PICTURE";
    case HWPTAG_SHAPE_COMPONENT_CONTAINER:
        return "HWPTAG_SHAPE_COMPONENT_CONTAINER";
    case HWPTAG_CTRL_DATA:
        return "HWPTAG_CTRL_DATA";
    case HWPTAG_EQEDIT:
        return "HWPTAG_EQEDIT";
    case HWPTAG_SHAPE_COMPONENT_TEXTART:
        return "HWPTAG_SHAPE_COMPONENT_TEXTART";
    case HWPTAG_FORM_OBJECT:
        return "HWPTAG_FORM_OBJECT";
    case HWPTAG_MEMO_SHAPE:
        return "HWPTAG_MEMO_SHAPE";
    case HWPTAG_MEMO_LIST:
        return "HWPTAG_MEMO_LIST";
    case HWPTAG_CHART_DATA:
        return "HWPTAG_CHART_DATA";
    case HWPTAG_SHAPE_COMPONENT_UNKNOWN:
        return "HWPTAG_SHAPE_COMPONENT_UNKNOWN";
    }
    return "(UNIDENTIFIED)";
}

char *hwp5_ctrl_name(DWORD ctrl_id)
{
    static char unknown[26];
    char a, b, c, d;
    switch (ctrl_id) {
    case HWPCTRL_TABLE:
        return "table";
    case HWPCTRL_LINE:
        return "line";
    case HWPCTRL_RECTANGULAR:
        return "rectangular";
    case HWPCTRL_ELLIPSE:
        return "ellipse";
    case HWPCTRL_ARC:
        return "arc";
    case HWPCTRL_POLYGON:
        return "polygon";
    case HWPCTRL_CURVE:
        return "curve";
    case HWPCTRL_EQUATION:
        return "equation";
    case HWPCTRL_PICTURE:
        return "picture";
    case HWPCTRL_OLE:
        return "ole";
    case HWPCTRL_CONTAINER:
        return "container";
    case HWPCTRL_SECTION_DEFINITION:
        return "section_definition";
    case HWPCTRL_COLUME_DEFINITION:
        return "colume_definition";
    case HWPCTRL_HEADER:
        return "header";
    case HWPCTRL_FOOTER:
        return "footer";
    case HWPCTRL_FOOT_NOTE:
        return "foot_note";
    case HWPCTRL_END_NOTE:
        return "end_note";
    case HWPCTRL_AUTO_NUMBER:
        return "auto_number";
    case HWPCTRL_NEW_NUMBER:
        return "new_number";
    case HWPCTRL_PAGE_HIDE:
        return "page_hide";
    case HWPCTRL_PAGE_ODD_EVEN_CONTROL:
        return "page_odd_even_control";
    case HWPCTRL_PAGE_NUMBER_POSITION:
        return "page_number_position";
    case HWPCTRL_INDEX_MARK:
        return "index_mark";
    case HWPCTRL_BOOKMARK:
        return "bookmark";
    case HWPCTRL_CHARACTER_OVERLAP:
        return "character_overlap";
    case HWPCTRL_COMMENT:
        return "comment";
    case HWPCTRL_HIDDEN_DESCRIPTION:
        return "hidden_description";
    case HWPCTRL_FLD_UNKNOWN:
        return "fld_unknown";
    case HWPCTRL_FLD_DATE:
        return "fld_date";
    case HWPCTRL_FLD_DOCDATE:
        return "fld_docdate";
    case HWPCTRL_FLD_PATH:
        return "fld_path";
    case HWPCTRL_FLD_BOOKMARK:
        return "fld_bookmark";
    case HWPCTRL_FLD_MAILMERGE:
        return "fld_mailmerge";
    case HWPCTRL_FLD_CROSSREF:
        return "fld_crossref";
    case HWPCTRL_FLD_FORMULA:
        return "fld_formula";
    case HWPCTRL_FLD_CLICKHERE:
        return "fld_clickhere";
    case HWPCTRL_FLD_SUMMARY:
        return "fld_summary";
    case HWPCTRL_FLD_USERINFO:
        return "fld_userinfo";
    case HWPCTRL_FLD_HYPERLINK:
        return "fld_hyperlink";
    case HWPCTRL_FLD_REV_SIGN:
        return "fld_rev_sign";
    case HWPCTRL_FLD_REV_DELETE:
        return "fld_rev_delete";
    case HWPCTRL_FLD_REV_ATTACH:
        return "fld_rev_attach";
    case HWPCTRL_FLD_REV_CLIPPING:
        return "fld_rev_clipping";
    case HWPCTRL_FLD_REV_SAWTOOTH:
        return "fld_rev_sawtooth";
    case HWPCTRL_FLD_REV_THINKING:
        return "fld_rev_thinking";
    case HWPCTRL_FLD_REV_PRAISE:
        return "fld_rev_praise";
    case HWPCTRL_FLD_REV_LINE:
        return "fld_rev_line";
    case HWPCTRL_FLD_REV_SIMPLECHANGE:
        return "fld_rev_simplechange";
    case HWPCTRL_FLD_REV_HYPERLINK:
        return "fld_rev_hyperlink";
    case HWPCTRL_FLD_REV_LINEATTACH:
        return "fld_rev_lineattach";
    case HWPCTRL_FLD_REV_LINELINK:
        return "fld_rev_linelink";
    case HWPCTRL_FLD_REV_LINETRANSFER:
        return "fld_rev_linetransfer";
    case HWPCTRL_FLD_REV_RIGHTMOVE:
        return "fld_rev_rightmove";
    case HWPCTRL_FLD_REV_LEFTMOVE:
        return "fld_rev_leftmove";
    case HWPCTRL_FLD_REV_TRANSFER:
        return "fld_rev_transfer";
    case HWPCTRL_FLD_REV_SIMPLEINSERT:
        return "fld_rev_simpleinsert";
    case HWPCTRL_FLD_REV_SPLIT:
        return "fld_rev_split";
    case HWPCTRL_FLD_REV_CHANGE:
        return "fld_rev_change";
    case HWPCTRL_FLD_MEMO:
        return "fld_memo";
    case HWPCTRL_FLD_PRIVATE_INFO_SECURITY:
        return "fld_private_info_security";
    }
    a = (unsigned int) ctrl_id >> 24;
    b = (unsigned int) ctrl_id >> 16;
    c = (unsigned int) ctrl_id >> 8;
    d = ctrl_id;
    snprintf(unknown, 26, "4chid('%c', '%c', '%c', '%c')", a, b, c, d);
    return unknown;
}

void hwp5_print_hdr_list(const char *name, int size, HDRSTREAM *h)
{
    int i;
    HDR *hdr;

    printf("\n%s (%d bytes) records list\n\n", name, size);

    i = 0;
    while (NULL != (hdr = hwp5_alloc_hdr(HDR_NEXT_RECORD, h))) {
        ++i;
        hwp5_print_hdr(i, hdr);
        hwp5_free_hdr(hdr);
    }
}

void hwp5_print_hdr_table(const char *name, int size, HDRSTREAM *h)
{
    int i;
    HDR *hdr;

    printf("\n%s (%d bytes) records table\n", name, size);
    printf("\n  no    id   lev   size       tag name\n");
    i = 0;
    while (NULL != (hdr = hwp5_alloc_hdr(HDR_NEXT_RECORD, h))) {
        ++i;
        hwp5_print_hdr_row(i, hdr);
        hwp5_free_hdr(hdr);
    }
}

void hwp5_print_hdr(int n, HDR *hdr)
{
    int id, lev, size;
    char *name;
    id = hdr->header.tag_id;
    lev = hdr->header.level;
    name = hdr_tag_name(id);
    size = hdr->size;
    printf(" #%d %s (id:%d lev:%d size:%d)\n",
               n, name, id,   lev,   size);
    if (id == HWPTAG_CTRL_HEADER)
        printf(" ctrl name: %s\n", hwp5_ctrl_name(*(DWORD *)(hdr->data)));
    print_byte_seq(hdr->data, hdr->size, NUMBERING_MODE);
    if (id == HWPTAG_PARA_TEXT) {
        printf(" text:\n");
        hwp5_print_para_text(hdr);
    }
    putchar('\n');
}

void hwp5_print_hdr_row(int n, HDR *hdr)
{
    int id, lev, size;
    char *name;
    if (hdr == NULL)
        return;
    id = hdr->header.tag_id;
    lev = hdr->header.level;
    name = hdr_tag_name(id);
    size = hdr->size;
    printf(" %3d  %4d  %4d   %4d   %s",
              n,  id,  lev, size, name);
    if (id == HWPTAG_CTRL_HEADER)
        printf(" (%s)\n", hwp5_ctrl_name(*(DWORD *)(hdr->data)));
    else
        putchar('\n');
}

void hwp5_print_hdrstream(const char *path, HWP5FILE *hfp)
{
    HDRSTREAM *h;
    int size;
    h = hwp5_alloc_hdrstream(path, hfp);
    if (h == NULL)
        return;

    size = h->stream->data_size;
    hwp5_print_hdr_table(path, size, h);
    h->offset = 0; /* rewind! */
    hwp5_print_hdr_list(path, size, h);
    hwp5_free_hdrstream(h);
}

void hwp5_print_header(HWP5FILE *hfp)
{
    char ver_str[16];
    hwp5_get_version_string(ver_str, 16, hfp);
    printf("\nFile header:\n");
    printf("    Signature : %s\n", hfp->header.signature);
    printf("    Version   : %s\n", ver_str);
    printf("    Attribute : ");
    hwp5_print_header_attribute(hfp->header.attribute);
    printf("\n");
}

void hwp5_print_header_attribute(DWORD attr)
{
    char *attr_str[] = {
        "Compress", "Password", "Distribute", "Script", "DRM",
        "XML_Template", "Document_History_Control", "Electronic_Signature",
        "Public_Key_Certificate", "Electronic_Signature_Reserve",
        "Public_Key_Certificate_&_DRM", "CCL"
    };
    int n = sizeof(attr_str) / sizeof(attr_str[0]);
    int i;

    for (i = 0; i < n; ++i) {
        if ((attr & 1) == 1)
            printf("%s ", attr_str[i]);
        attr = (unsigned) attr >> 1;
    }
}

void hwp5_print_summary(HWP5FILE *hfp)
{
    printf("\nHWP summary information:\n");
    printf("    Title   : %s\n", hfp->summary.title);
    printf("    Subject : %s\n", hfp->summary.subject);
    printf("    Author  : %s\n", hfp->summary.author);
    printf("    Date    : %s\n", hfp->summary.date);
    printf("    Keyword : %s\n", hfp->summary.keyword);
    printf("    Others  : %s\n", hfp->summary.others);
}


void hwp5_print_prvtext(HWP5FILE *hfp)
{
    if (hfp->prvtext == NULL)
        return;
    printf("\nPreview text:\n");
    printf("%s\n", hfp->prvtext);
}

void hwp5_print_section(HWP5FILE *hfp)
{
    LNode *n;
    LNode *base = cff_make_dir_list(hfp->cfp);
    CffDirLNode *ent;
    if (base == NULL)
        return;
    for (n = base->next; n != NULL; n = n->next) {
        ent = n->data;
        if (NULL == ent || ent->name == NULL)
            continue;
        if (NULL != strstr(ent->name, "Section"))
            hwp5_print_hdrstream(ent->path, hfp);
    }
    cff_destroy_dir_list(base);
}

void hwp5_print_all(HWP5FILE *hfp)
{
    hwp5_print_header(hfp);
    hwp5_print_summary(hfp);
    hwp5_print_prvtext(hfp);
    if (IS_OPTION(TEXT_WITHOUT_HEX)) {
        hwp5_print_body_text(hfp);
    } else {
        hwp5_print_hdrstream("DocInfo", hfp);
        hwp5_print_section(hfp);
    }
}

/* path를 이용해 새 파일 이름을 만들고 표준출력을 파일로 돌린다.
   성공시 0 실패시 -1 반환한다.  */
static int redirect_hwp5dump_file(const char *path)
{
    return redirect_stdout_to_file(!IS_OPTION(OVERWRITE),
                                   path, "hwp5dump.txt");
}

void print_hwp5(const char *path, HWP5FILE *hfp)
{
    if (!IS_OPTION(STANDARD_OUTPUT))
        if (-1 == redirect_hwp5dump_file(path))
            return;

    printf("'%s'\n", path);
    hwp5_print_all(hfp);
}

int hwp5_file_dump(const char *path)
{
    HWP5FILE *hfp;

    if (path == NULL)
        return -1;

    hfp = hwp5_open(path, "rb");
    if (hfp == NULL) {
        fprintf(stderr, "Cannot open %s\n", path);
        return 1;
    }

    if (IS_OPTION(DESTINATION_DIRECTORY))
        if (-1 == make_dir_chdir(destination_directory)) {
            hwp5_close(hfp);
            return -1;
        }

    if (IS_OPTION(BINDATA_FILE))
        hwp5_extract_bindata_file(path, hfp);
    else
        print_hwp5(path, hfp);

    hwp5_close(hfp);
    return 0;
}

void help_msg()
{
    printf("Usage: hwp5dump [OPTION] FILE...\n"
           "Dump hwp 5.0 format FILEs\n\n"
           "Option:\n"
           "  -t              Write text without hexadecimal dump\n"
           "  -d DIRECTORY    Create output files in DIRECTORY\n"
           "  -p              Print to standard output\n"
           "  -o              Overwrite output files\n"
           "  -b              Extract files in BinData storage\n"
           "  --              Next arguments are not an option\n"
           "  --help\n"
           "  --version\n\n"
           "Report bugs to <http://kldp.net/projects/cffdump/>\n");
#ifdef IN_WINDOWS
    printf("\nIn Windows, you can drag and drop hwp files "
           "on the program icon\n");
    press_any_key();
#endif
    exit(0);
}

void version_msg()
{
    printf(PROGRAM_NAME_AND_VERSION"\n"
           "Written by Young ho Park\n"
           "This program is following HWP Document File Formats Open Policy of Hancom\n");

#ifdef IN_WINDOWS
    press_any_key();
#endif
    exit(0);
}

/* 배열에서 wc와 같은 요소를 제거한다.  WCHAR 배열이지만
   중간중간 널 문자가 끼어있기 때문에 배열요소개수가 필요하다.  */
void del_wchar_in_ar(WCHAR wc, WCHAR *array, unsigned *elem_num)
{
    unsigned i, j;
    WCHAR *a = array;

    i = j = 0;
    while (i < *elem_num)
        if (a[i] != wc) {
            a[j] = a[i];
            j++;
            i++;
        } else {
            i++;
        }

    *elem_num = j;
}

/* ar[index]부터 배열요소를 del_num개 지운다.  */
void del_any_wchar_in_ar(unsigned index,
                         unsigned del_num,
                         WCHAR *ar,
                         unsigned *elem_num)
{
    unsigned copy_num;
    if (index + del_num >= *elem_num) {
        *elem_num = index;
        return;
    }
    copy_num = *elem_num - index - del_num;
    memmove(&ar[index], &ar[index + del_num], copy_num * sizeof(WCHAR));
    *elem_num -= del_num;
}

/* 읽지 않을 단순 특수 문자를 삭제한다.  */
void del_special_wchar(WCHAR *ar, unsigned *elem_num)
{
    /* 단순특수문자 중 0(빈공간), 10(줄끝), 13(문단끝),
       24(하이픈), 30, 31(빈칸)은 남겨둔다.  */
    static const WCHAR del[] = {
        HWPCHAR_CHR_RESERVED25,
        HWPCHAR_CHR_RESERVED26,
        HWPCHAR_CHR_RESERVED27,
        HWPCHAR_CHR_RESERVED28,
        HWPCHAR_CHR_RESERVED29
    };
    static const unsigned n = sizeof(del) / sizeof(del[0]);

    int i;
    for (i = 0; i < n; ++i)
        del_wchar_in_ar(del[i], ar, elem_num);
}

/* extended, inline 특수문자는 같은 방식으로 지울 수 있다.  */
void del_extended_inline(WCHAR *ar, unsigned *elem_num)
{
    static const WCHAR del[] = {
    /* extended 글자는 다 지운다.  */
        HWPCHAR_EXT_RESERVED1,
        HWPCHAR_EXT_SECTION_COLUMN,
        HWPCHAR_EXT_FIELD_START,
        HWPCHAR_EXT_DRAW_TABLE,
        HWPCHAR_EXT_RESERVED12,
        HWPCHAR_EXT_RESERVED14,
        HWPCHAR_EXT_HIDDEN_DESCRIPTION,
        HWPCHAR_EXT_HEAD_TAIL,
        HWPCHAR_EXT_FOOTNOTE,
        HWPCHAR_EXT_AUTO_NUMBER,
        HWPCHAR_EXT_PAGE_CONTROL,
        HWPCHAR_EXT_INDEX_MARK,
        HWPCHAR_EXT_OVERLAP,
    /* inline 글자는 9(탭)만 남겨둔다.  */
        HWPCHAR_INL_FIELD_END,
        HWPCHAR_INL_RESERVED5,
        HWPCHAR_INL_RESERVED6,
        HWPCHAR_INL_RESERVED7,
        HWPCHAR_INL_TITLE_MARK,
        HWPCHAR_INL_RESERVED19,
        HWPCHAR_INL_RESERVED20
    };
    static const unsigned n = sizeof(del) / sizeof(del[0]);

    unsigned i, j;
    for (i = 0; i < n; ++i)
        for (j = 0; j < *elem_num; ++j)
            if (del[i] == ar[j])
                del_any_wchar_in_ar(j--, 8, ar, elem_num);
}

/* 탭 인라인 특수 문자를 일반 문자로 바꾼다.  */
void tr_inline_tab(WCHAR *ar, unsigned *elem_num)
{
    unsigned i;
    for (i = 0; i < *elem_num; ++i)
        if (ar[i] == HWPCHAR_INL_TAB) {
            del_any_wchar_in_ar(i, 7, ar, elem_num);
            ar[i] = '\t';
        }
}

/* 단순 특수 문자를 일반 문자로 바꾼다.  */
void tr_special_wchar(WCHAR *ar, unsigned elem_num)
{
    unsigned i;
    for (i = 0; i < elem_num; ++i)
        switch (ar[i]) {
        case HWPCHAR_CHR_LINE_BREAK:
        case HWPCHAR_CHR_PARA_BREAK:
            ar[i] = '\n';
            break;
        case HWPCHAR_CHR_HYPHEN:
            ar[i] = '-';
            break;
        case HWPCHAR_CHR_BUNCH_SPACE:
        case HWPCHAR_CHR_FIXED_SPACE:
            ar[i] = ' ';
            break;
        default:
            break;
        }
}

/* 문단 데이터에서 특수 문자들을 지우거나 변환하고
   컴파일 매크로에 따라 utf-8이나 windows c locale
   텍스트로 변환한다.  미리보기와 마찬가지로 본문에도
   따로 마지막을 나타내는 표식이 없다.  */
void hwp5_print_para_text(HDR *p)
{
    WCHAR *ar = p->data;
    unsigned num = p->size / sizeof(WCHAR);
    char *para_text;

    del_extended_inline(ar, &num);
    del_special_wchar(ar, &num);
    tr_special_wchar(ar, num);
    tr_inline_tab(ar, &num);
    del_wchar_in_ar(0, ar, &num);

    para_text = strdup_from_utf16(ar, num);
    if (para_text == NULL)
        return;

    printf("%s\n", para_text);
    free(para_text);
}

void hwp5_print_section_text(HDRSTREAM *s)
{
    HDR *para;
    while (NULL != (para = hwp5_alloc_hdr(HWPTAG_PARA_TEXT, s))) {
        hwp5_print_para_text(para);
        hwp5_free_hdr(para);
    }
}

void hwp5_print_body_text(HWP5FILE *hfp)
{
    HDRSTREAM *section;
    LNode *n;
    LNode *base;
    CffDirLNode *ent;

    base = cff_make_dir_list(hfp->cfp);
    if (base == NULL)
        return;

    printf("\nBody text:\n");
    for (n = base->next; n != NULL; n = n->next) {
        ent = n->data;
        if (NULL == ent || ent->name == NULL)
            continue;
        if (NULL != strstr(ent->name, "Section")) {
            section = hwp5_alloc_hdrstream(ent->path, hfp);
            if (section == NULL)
                continue;
            hwp5_print_section_text(section);
            hwp5_free_hdrstream(section);
        }
    }
    cff_destroy_dir_list(base);
}

void hwp5_split_bindata_storage(CffDirLNode *bindata,
                                LNode *base,
                                const char *path,
                                HWP5FILE *hfp)
{
    CffStreamFilter filter = NULL;
    if (path == NULL || bindata == NULL || base == NULL || hfp == NULL)
        return;
    if (HWP_IS_COMPRESSED(hfp))
        filter = hwp5_decompress;

    cff_make_split_stream_storage(bindata, base, hfp->cfp,
                                  filter, !IS_OPTION(OVERWRITE), path);
}

void hwp5_extract_bindata_file(const char *path, HWP5FILE *hfp)
{
    LNode *b;
    LNode *n;
    CffDirLNode *d;

    if (hfp == NULL || hfp->cfp == NULL)
        return;

    b = cff_make_dir_list(hfp->cfp);
    if (b == NULL)
        return;

    for (n = b->next; n != NULL; n = n->next) {
        d = n->data;
        if (d == NULL || d->name == NULL)
            continue;
        if (0 == strcmp(d->name, "BinData")) {
            hwp5_split_bindata_storage(d, b, path, hfp);
            break;
        }
    }

    cff_destroy_dir_list(b);
}


void read_option(int *file_start, int argc, char **argv)
{
    int i;
    char opt;

    for (i = 1; i < argc; ++i) {
        if (argv[i][0] == '-') {
            opt = argv[i][1];
            switch (opt) {
            case 'd':
                option |= DESTINATION_DIRECTORY;
                ++i;
                if (i + 1 >= argc) {
                    fputs("Usage: hwp5dump -d DIRECTORY FILE...\n", stderr);
                    exit(1);
                }
                destination_directory = argv[i];
                break;
            case 'b':
                option |= BINDATA_FILE;
                break;
            case 'p':
                option |= STANDARD_OUTPUT;
                break;
            case 'o':
                option |= OVERWRITE;
                break;
            case 't':
                option |= TEXT_WITHOUT_HEX;
                break;
            case '-':
                ++i;
                goto option_finish;
            default:
                fprintf(stderr, "Unknown option: %c\n", opt);
                exit(1);
                break;
            }
        } else {
            break;
        }
    }

option_finish:
    if (i >= argc) {
        fputs("Usage: hwp5dump [OPTION] FILE...\n", stderr);
        exit(1);
    }
    *file_start = i;
}

int main(int argc, char **argv)
{
    int i;
    int err;
    int file_start = 1;
    char *start_dir;

    if (argc <= 1 || 0 == strcmp("--help", argv[1]))
        help_msg();
    else if (0 == strcmp("--version", argv[1]))
        version_msg();

    read_option(&file_start, argc, argv);

    start_dir = strdup_pwd();
    if (start_dir == NULL)
        return 1;

    for (i = file_start; i < argc; ++i) {
        if (-1 == hwp5_file_dump(argv[i]))
            break;
        if (-1 == chdir(start_dir)) {
            err = errno;
            file_err_msg(err, start_dir);
            break;
        }
    }

    free(start_dir);
#ifdef IN_WINDOWS
    press_any_key();
#endif

    return 0;
}
