/* hwp5type.h

   hwp5dump는 한글과컴퓨터의 한글 문서 파일 구조 정의
   공개 정책을 따르는 2차 저작물입니다.  */

#ifndef HWP5TYPE_H
#define HWP5TYPE_H

#include <inttypes.h>
#include "cffread.h"

/* HWP 파일 헤더 구조체 속성 파악 */
#define HWP_IS_COMPRESSED(HFP) ((HFP)->header.attribute & 0x1)
#define HWP_IS_LOCKED(HFP)     ((HFP)->header.attribute & 0x2)

#define HDR_NEXT_RECORD (-1)
/* HWP 데이터 레코드 태그 ID 시작 */
#define HWPTAG_BEGIN 16

/* 문서 정보 태그 ID */
#define HWPTAG_DOCUMENT_PROPERTIES  (HWPTAG_BEGIN)
#define HWPTAG_ID_MAPPINGS          (HWPTAG_BEGIN + 1)
#define HWPTAG_BIN_DATA             (HWPTAG_BEGIN + 2)
#define HWPTAG_FACE_NAME            (HWPTAG_BEGIN + 3)
#define HWPTAG_BORDER_FILL          (HWPTAG_BEGIN + 4)
#define HWPTAG_CHAR_SHAPE           (HWPTAG_BEGIN + 5)
#define HWPTAG_TAB_DEF              (HWPTAG_BEGIN + 6)
#define HWPTAG_NUMBERING            (HWPTAG_BEGIN + 7)
#define HWPTAG_BULLET               (HWPTAG_BEGIN + 8)
#define HWPTAG_PARA_SHAPE           (HWPTAG_BEGIN + 9)
#define HWPTAG_STYLE                (HWPTAG_BEGIN + 10)
#define HWPTAG_DOC_DATA             (HWPTAG_BEGIN + 11)
#define HWPTAG_DISTRIBUTE_DOC_DATA  (HWPTAG_BEGIN + 12)
#define HWPTAG_COMPATIBLE_DOCUMENT  (HWPTAG_BEGIN + 14)
#define HWPTAG_LAYOUT_COMPATIBILITY (HWPTAG_BEGIN + 15)
#define HWPTAG_FORBIDDEN_CHAR       (HWPTAG_BEGIN + 78)

/* 본문 태그 ID */
#define HWPTAG_PARA_HEADER               (HWPTAG_BEGIN + 50)
#define HWPTAG_PARA_TEXT                 (HWPTAG_BEGIN + 51)
#define HWPTAG_PARA_CHAR_SHAPE           (HWPTAG_BEGIN + 52)
#define HWPTAG_PARA_LINE_SEG             (HWPTAG_BEGIN + 53)
#define HWPTAG_PARA_RANGE_TAG            (HWPTAG_BEGIN + 54)
#define HWPTAG_CTRL_HEADER               (HWPTAG_BEGIN + 55)
#define HWPTAG_LIST_HEADER               (HWPTAG_BEGIN + 56)
#define HWPTAG_PAGE_DEF                  (HWPTAG_BEGIN + 57)
#define HWPTAG_FOOTNOTE_SHAPE            (HWPTAG_BEGIN + 58)
#define HWPTAG_PAGE_BORDER_FILL          (HWPTAG_BEGIN + 59)
#define HWPTAG_SHAPE_COMPONENT           (HWPTAG_BEGIN + 60)
#define HWPTAG_TABLE                     (HWPTAG_BEGIN + 61)
#define HWPTAG_SHAPE_COMPONENT_LINE      (HWPTAG_BEGIN + 62)
#define HWPTAG_SHAPE_COMPONENT_RECTANGLE (HWPTAG_BEGIN + 63)
#define HWPTAG_SHAPE_COMPONENT_ELLIPSE   (HWPTAG_BEGIN + 64)
#define HWPTAG_SHAPE_COMPONENT_ARC       (HWPTAG_BEGIN + 65)
#define HWPTAG_SHAPE_COMPONENT_POLYGON   (HWPTAG_BEGIN + 66)
#define HWPTAG_SHAPE_COMPONENT_CURVE     (HWPTAG_BEGIN + 67)
#define HWPTAG_SHAPE_COMPONENT_OLE       (HWPTAG_BEGIN + 68)
#define HWPTAG_SHAPE_COMPONENT_PICTURE   (HWPTAG_BEGIN + 69)
#define HWPTAG_SHAPE_COMPONENT_CONTAINER (HWPTAG_BEGIN + 70)
#define HWPTAG_CTRL_DATA                 (HWPTAG_BEGIN + 71)
#define HWPTAG_EQEDIT                    (HWPTAG_BEGIN + 72)
#define HWPTAG_SHAPE_COMPONENT_TEXTART   (HWPTAG_BEGIN + 74)
#define HWPTAG_FORM_OBJECT               (HWPTAG_BEGIN + 75)
#define HWPTAG_MEMO_SHAPE                (HWPTAG_BEGIN + 76)
#define HWPTAG_MEMO_LIST                 (HWPTAG_BEGIN + 77)
#define HWPTAG_CHART_DATA                (HWPTAG_BEGIN + 79)
#define HWPTAG_SHAPE_COMPONENT_UNKNOWN   (HWPTAG_BEGIN + 99)

/* HWP 문자 코드 */
#define HWPCHAR_CHR_UNUSABLE            0
#define HWPCHAR_EXT_RESERVED1           1
#define HWPCHAR_EXT_SECTION_COLUMN      2
#define HWPCHAR_EXT_FIELD_START         3
#define HWPCHAR_INL_FIELD_END           4
#define HWPCHAR_INL_RESERVED5           5
#define HWPCHAR_INL_RESERVED6           6
#define HWPCHAR_INL_RESERVED7           7
#define HWPCHAR_INL_TITLE_MARK          8
#define HWPCHAR_INL_TAB                 9
#define HWPCHAR_CHR_LINE_BREAK          10
#define HWPCHAR_EXT_DRAW_TABLE          11
#define HWPCHAR_EXT_RESERVED12          12
#define HWPCHAR_CHR_PARA_BREAK          13
#define HWPCHAR_EXT_RESERVED14          14
#define HWPCHAR_EXT_HIDDEN_DESCRIPTION  15
#define HWPCHAR_EXT_HEAD_TAIL           16
#define HWPCHAR_EXT_FOOTNOTE            17
#define HWPCHAR_EXT_AUTO_NUMBER         18
#define HWPCHAR_INL_RESERVED19          19
#define HWPCHAR_INL_RESERVED20          20
#define HWPCHAR_EXT_PAGE_CONTROL        21
#define HWPCHAR_EXT_INDEX_MARK          22
#define HWPCHAR_EXT_OVERLAP             23
#define HWPCHAR_CHR_HYPHEN              24
#define HWPCHAR_CHR_RESERVED25          25
#define HWPCHAR_CHR_RESERVED26          26
#define HWPCHAR_CHR_RESERVED27          27
#define HWPCHAR_CHR_RESERVED28          28
#define HWPCHAR_CHR_RESERVED29          29
#define HWPCHAR_CHR_BUNCH_SPACE         30
#define HWPCHAR_CHR_FIXED_SPACE         31

#define MAKE_4CHID(a, b, c, d) \
    (((a) << 24) | ((b) << 16) | ((c) << 8) | (d))
/* 컨트롤 아이디 */
#define HWPCTRL_TABLE                       MAKE_4CHID('t', 'b', 'l', ' ')
#define HWPCTRL_LINE                        MAKE_4CHID('$', 'l', 'i', 'n')
#define HWPCTRL_RECTANGULAR                 MAKE_4CHID('$', 'r', 'e', 'c')
#define HWPCTRL_ELLIPSE                     MAKE_4CHID('$', 'e', 'l', 'l')
#define HWPCTRL_ARC                         MAKE_4CHID('$', 'a', 'r', 'c')
#define HWPCTRL_POLYGON                     MAKE_4CHID('$', 'p', 'o', 'l')
#define HWPCTRL_CURVE                       MAKE_4CHID('$', 'c', 'u', 'r')
#define HWPCTRL_EQUATION                    MAKE_4CHID('e', 'q', 'e', 'd')
#define HWPCTRL_PICTURE                     MAKE_4CHID('$', 'p', 'i', 'c')
#define HWPCTRL_OLE                         MAKE_4CHID('$', 'o', 'l', 'e')
#define HWPCTRL_CONTAINER                   MAKE_4CHID('$', 'c', 'o', 'n')
#define HWPCTRL_SECTION_DEFINITION          MAKE_4CHID('s', 'e', 'c', 'd')
#define HWPCTRL_COLUME_DEFINITION           MAKE_4CHID('c', 'o', 'l', 'd')
#define HWPCTRL_HEADER                      MAKE_4CHID('h', 'e', 'a', 'd')
#define HWPCTRL_FOOTER                      MAKE_4CHID('f', 'o', 'o', 't')
#define HWPCTRL_FOOT_NOTE                   MAKE_4CHID('f', 'n', ' ', ' ')
#define HWPCTRL_END_NOTE                    MAKE_4CHID('e', 'n', ' ', ' ')
#define HWPCTRL_AUTO_NUMBER                 MAKE_4CHID('a', 't', 'n', 'o')
#define HWPCTRL_NEW_NUMBER                  MAKE_4CHID('n', 'w', 'n', 'o')
#define HWPCTRL_PAGE_HIDE                   MAKE_4CHID('p', 'g', 'h', 'd')
#define HWPCTRL_PAGE_ODD_EVEN_CONTROL       MAKE_4CHID('p', 'g', 'c', 't')
#define HWPCTRL_PAGE_NUMBER_POSITION        MAKE_4CHID('p', 'g', 'n', 'p')
#define HWPCTRL_INDEX_MARK                  MAKE_4CHID('i', 'd', 'x', 'm')
#define HWPCTRL_BOOKMARK                    MAKE_4CHID('b', 'o', 'k', 'm')
#define HWPCTRL_CHARACTER_OVERLAP           MAKE_4CHID('t', 'c', 'p', 's')
#define HWPCTRL_COMMENT                     MAKE_4CHID('t', 'd', 'u', 't')
#define HWPCTRL_HIDDEN_DESCRIPTION          MAKE_4CHID('t', 'c', 'm', 't')
#define HWPCTRL_FLD_UNKNOWN                 MAKE_4CHID('%', 'u', 'n', 'k')
#define HWPCTRL_FLD_DATE                    MAKE_4CHID('%', 'd', 't', 'e')
#define HWPCTRL_FLD_DOCDATE                 MAKE_4CHID('%', 'd', 'd', 't')
#define HWPCTRL_FLD_PATH                    MAKE_4CHID('%', 'p', 'a', 't')
#define HWPCTRL_FLD_BOOKMARK                MAKE_4CHID('%', 'b', 'm', 'k')
#define HWPCTRL_FLD_MAILMERGE               MAKE_4CHID('%', 'm', 'm', 'g')
#define HWPCTRL_FLD_CROSSREF                MAKE_4CHID('%', 'x', 'r', 'f')
#define HWPCTRL_FLD_FORMULA                 MAKE_4CHID('%', 'f', 'm', 'u')
#define HWPCTRL_FLD_CLICKHERE               MAKE_4CHID('%', 'c', 'l', 'k')
#define HWPCTRL_FLD_SUMMARY                 MAKE_4CHID('%', 's', 'm', 'r')
#define HWPCTRL_FLD_USERINFO                MAKE_4CHID('%', 'u', 's', 'r')
#define HWPCTRL_FLD_HYPERLINK               MAKE_4CHID('%', 'h', 'l', 'k')
#define HWPCTRL_FLD_REV_SIGN                MAKE_4CHID('%', 's', 'i', 'g')
#define HWPCTRL_FLD_REV_DELETE              MAKE_4CHID('%', '%', '*', 'd')
#define HWPCTRL_FLD_REV_ATTACH              MAKE_4CHID('%', '%', '*', 'a')
#define HWPCTRL_FLD_REV_CLIPPING            MAKE_4CHID('%', '%', '*', 'C')
#define HWPCTRL_FLD_REV_SAWTOOTH            MAKE_4CHID('%', '%', '*', 'S')
#define HWPCTRL_FLD_REV_THINKING            MAKE_4CHID('%', '%', '*', 'T')
#define HWPCTRL_FLD_REV_PRAISE              MAKE_4CHID('%', '%', '*', 'P')
#define HWPCTRL_FLD_REV_LINE                MAKE_4CHID('%', '%', '*', 'L')
#define HWPCTRL_FLD_REV_SIMPLECHANGE        MAKE_4CHID('%', '%', '*', 'c')
#define HWPCTRL_FLD_REV_HYPERLINK           MAKE_4CHID('%', '%', '*', 'h')
#define HWPCTRL_FLD_REV_LINEATTACH          MAKE_4CHID('%', '%', '*', 'A')
#define HWPCTRL_FLD_REV_LINELINK            MAKE_4CHID('%', '%', '*', 'i')
#define HWPCTRL_FLD_REV_LINETRANSFER        MAKE_4CHID('%', '%', '*', 't')
#define HWPCTRL_FLD_REV_RIGHTMOVE           MAKE_4CHID('%', '%', '*', 'r')
#define HWPCTRL_FLD_REV_LEFTMOVE            MAKE_4CHID('%', '%', '*', 'l')
#define HWPCTRL_FLD_REV_TRANSFER            MAKE_4CHID('%', '%', '*', 'n')
#define HWPCTRL_FLD_REV_SIMPLEINSERT        MAKE_4CHID('%', '%', '*', 'e')
#define HWPCTRL_FLD_REV_SPLIT               MAKE_4CHID('%', 's', 'p', 'l')
#define HWPCTRL_FLD_REV_CHANGE              MAKE_4CHID('%', '%', 'm', 'r')
#define HWPCTRL_FLD_MEMO                    MAKE_4CHID('%', '%', 'm', 'e')
#define HWPCTRL_FLD_PRIVATE_INFO_SECURITY   MAKE_4CHID('%', 'c', 'p', 'r')

/* 자료형 */
typedef uint8_t     BYTE;
typedef uint16_t    WORD;
typedef uint32_t    DWORD;
typedef uint16_t    WCHAR;
typedef uint32_t    HWPUNIT;
typedef int32_t     SHWPUNIT;
typedef uint8_t     UINT8;
typedef uint16_t    UINT16;
typedef uint32_t    UINT32, UINT;
typedef int8_t      INT8;
typedef int16_t     INT16;
typedef int32_t     INT32;
typedef INT16       HWPUNIT16;
typedef uint32_t    COLORREF;

/* HWP 파일 헤더 */
typedef struct {
    BYTE  signature[32];
    DWORD version;
    DWORD attribute;
    BYTE  reserved[216];
} HWPHEADER;

/* 문서 요약 */
typedef struct {
    char *title;
    char *subject;
    char *author;
    char *date;
    char *keyword;
    char *others;
} HWPSUMMARY;

/* 데이터 레코드 헤더 */
typedef struct {
    DWORD tag_id : 10;
    DWORD level  : 10;
    DWORD size   : 12;
} HDRHEADER;

/* 데이터 레코드 */
typedef struct {
    HDRHEADER header;
    DWORD     size;
    void     *data;
} HDR;

/* 데이터 레코드용 스트림 */
typedef struct {
    CFFSTREAM *stream;
    unsigned   offset;
} HDRSTREAM;

/* HWP5FILE 구조체 */
struct hwp5_hwpfile_struct {
    COMPOUNDFILE *cfp;
    HWPHEADER     header;
    HWPSUMMARY    summary;
    char         *prvtext;
};

#endif /* HWP5TYPE_H */
