#include <stdlib.h>
#include "unicode_encoding.h"
#include "unicode_string.h"

void ue_utf16_to_ucs4(const utf16char *src,
                      unsigned int     src_num,
                      ucs4char        *dest,
                      unsigned int     dest_num)
{
    unsigned int s_cnt, d_cnt;
    s_cnt = d_cnt = 0;

    while (*src != '\0') {
        ++d_cnt;
        if (dest_num != 0 && d_cnt > dest_num) {
            --dest;
            break;
        }
        ++s_cnt;
        if (src_num != 0 && s_cnt > src_num)
            break;
        if ((*src >= 0xD800) && (*src < 0xDC00)) {
            if ((*(src + 1) >= 0xDC00) && (*(src + 1) < 0xE000)) {
                ++s_cnt;
                if (src_num != 0 && s_cnt > src_num)
                    break;
                *dest = ((*src >> 6) & 0x000F) + 1;
                *dest <<= 16;
                *dest |= (*src & 0x003F) << 10;
                ++src;
                *dest |= *src & 0x03FF;
                ++src;
                ++dest;
            } else {
                --d_cnt;
                ++src;
            }
        } else {
            *dest = *src;
            ++dest;
            ++src;
        }
    }
    *dest = L'\0';
}


void ue_ucs4_to_utf16(const ucs4char *src,
                      unsigned int    src_num,
                      utf16char      *dest,
                      unsigned int    dest_num)
{
    unsigned int s_cnt, d_cnt;
    s_cnt = d_cnt = 0;

    while (*src != L'\0') {
        ++d_cnt;
        if (dest_num != 0 && d_cnt > dest_num) {
            --dest;
            break;
        }
        ++s_cnt;
        if (src_num != 0 && s_cnt > src_num)
            break;
        if ((*src >= 0x10000) && (*src < 0x110000)) {
            ++d_cnt;
            if (dest_num != 0 && d_cnt > dest_num) {
                --dest;
                break;
            }
            *dest = 0xD800 | (((*src & 0x1F0000) - 0x10000) >> 10);
            ++dest;
            *dest = 0xDC00 | (*src & 0x0003FF);
            ++dest;
            ++src;
        } else if (*src < 0x10000) {
            *dest = *src;
            ++dest;
            ++src;
        } else {
            --d_cnt;
            ++src;
        }
    }
    *dest = '\0';
}


void ue_utf8_to_ucs4(const utf8char *src,
                     unsigned int    src_num,
                     ucs4char       *dest,
                     unsigned int    dest_num)
{
    unsigned int s_cnt, d_cnt;
    s_cnt = d_cnt = 0;

    while (*src != '\0') {
        ++d_cnt;
        if (dest_num != 0 && d_cnt > dest_num) {
            --dest;
            break;
        }
        ++s_cnt;
        if (src_num != 0 && s_cnt > src_num)
            break;
        /* 1바이트짜리 utf-8 글자 */
        if ((*src & 0x80) == 0) {
            *dest = *src;
            ++dest;
            ++src;
        /* 2바이트짜리 utf-8 글자 */
        } else if ((*src & 0xE0) == 0xC0) {
            if ((*(src + 1) & 0xC0) == 0x80) {
                ++s_cnt;
                if (src_num != 0 && s_cnt > src_num)
                    break;
                *dest = (*src & 0x1F) << 6;
                ++src;
                *dest |= *src & 0x3F;
                ++src;
                ++dest;
            } else {
                --d_cnt;
                ++src;
            }
        /* 3바이트짜리 utf-8 글자 */
        } else if ((*src & 0xF0) == 0xE0) {
            if (((*(src + 1) & 0xC0) == 0x80) && ((*(src + 2) & 0xC0) == 0x80)) {
                s_cnt += 2;
                if (src_num != 0 && s_cnt > src_num)
                    break;
                *dest = (*src & 0x0F) << 12;
                ++src;
                *dest |= (*src & 0x3F) << 6;
                ++src;
                *dest |= *src & 0x3F;
                ++src;
                ++dest;
            } else {
                --d_cnt;
                ++src;
            }
        /* 4바이트짜리 utf-8 글자 */
        } else if ((*src & 0xF8) == 0xF0) {
            if (((*(src + 1) & 0xC0) == 0x80) && ((*(src + 2) & 0xC0) == 0x80) && ((*(src + 3) & 0xC0) == 0x80)) {
                s_cnt += 3;
                if (src_num != 0 && s_cnt > src_num)
                    break;
                *dest = (*src & 0x7) << 18;
                ++src;
                *dest |= (*src & 0x3F) << 12;
                ++src;
                *dest |= (*src & 0x3F) << 6;
                ++src;
                *dest |= *src & 0x3F;
                ++src;
                ++dest;
            } else {
                --d_cnt;
                ++src;
            }
        /* 5바이트 이상의 utf-8 글자는 준비 안함 */
        } else {
            --d_cnt;
            ++src;
        }
    }
    *dest = L'\0';
}

void ue_ucs4_to_utf8(const ucs4char *src,
                     unsigned int    src_num,
                     utf8char       *dest,
                     unsigned int    dest_num)
{
    unsigned int s_cnt, d_cnt;
    s_cnt = d_cnt = 0;

    while (*src != L'\0') {
        ++d_cnt;
        if (dest_num != 0 && d_cnt > dest_num) {
            --dest;
            break;
        }
        ++s_cnt;
        if (src_num != 0 && s_cnt > src_num)
            break;
        /* 1바이트 utf-8 문자로 */
        if (*src < 0x80) {
            *dest = *src;
            ++dest;
            ++src;
        /* 2바이트 utf-8 문자로 */
        } else if ((*src >= 0x80) && (*src < 0x0800)) {
            ++d_cnt;
            if (dest_num != 0 && d_cnt > dest_num) {
                --dest;
                break;
            }
            *dest = 0xC0 | ((*src >> 6) & 0x1F);
            ++dest;
            *dest = 0x80 | (*src & 0x3F);
            ++dest;
            ++src;
        /* 3바이트 utf-8 문자로 */
        } else if ((*src >= 0x0800) && (*src < 0x10000)) {
            d_cnt += 2;
            if (dest_num != 0 && d_cnt > dest_num) {
                --dest;
                break;
            }
            *dest = 0xE0 | ((*src >> 12) & 0xF);
            ++dest;
            *dest = 0x80 | ((*src >> 6) & 0x3F);
            ++dest;
            *dest = 0x80 | (*src & 0x3F);
            ++dest;
            ++src;
        /* 4바이트 utf-8 문자로 */
        } else if ((*src >= 0x10000) && (*src < 0x110000)) {
            d_cnt += 3;
            if (dest_num != 0 && d_cnt > dest_num) {
                --dest;
                break;
            }
            *dest = 0xF0 | ((*src >> 18) & 0x7);
            ++dest;
            *dest = 0x80 | ((*src >> 12) & 0x3F);
            ++dest;
            *dest = 0x80 | ((*src >> 6) & 0x3F);
            ++dest;
            *dest = 0x80 | (*src & 0x3F);
            ++dest;
            ++src;
        /* 5바이트 이상의 utf-8 문자는 준비 안함 */
        } else {
            --d_cnt;
            ++src;
        }
    }
    *dest = '\0';
}

void ue_utf16_to_utf8(const utf16char *src,
                      unsigned int     src_num,
                      utf8char        *dest,
                      unsigned int     dest_num)
{
    unsigned int ucs4_num = us_ucs4_len_of_utf16(src, src_num) + 1;
    ucs4char *medium = malloc(ucs4_num * sizeof(ucs4char));
    if (medium == NULL)
        return;
    ue_utf16_to_ucs4(src, src_num, medium, ucs4_num);
    ue_ucs4_to_utf8(medium, ucs4_num, dest, dest_num);
    free(medium);
}

void ue_utf8_to_utf16(const utf8char *src,
                      unsigned int    src_num,
                      utf16char      *dest,
                      unsigned int    dest_num)
{
    unsigned int ucs4_num = us_ucs4_len_of_utf8(src, src_num) + 1;
    ucs4char *medium = malloc(ucs4_num * sizeof(ucs4char));
    if (medium == NULL)
        return;
    ue_utf8_to_ucs4(src, src_num, medium, ucs4_num);
    ue_ucs4_to_utf16(medium, ucs4_num, dest, dest_num);
    free(medium);
}
