#include <string.h>
#include <stdlib.h>
#include "unicode_encoding.h"
#include "unicode_string.h"
#include "unicode_locale.h"

unsigned int us_ucs4_len_of_utf16(const utf16char *str, unsigned int num)
{
    unsigned int cnt, ucs4_len;
    cnt = ucs4_len = 0;

    while (*str != '\0') {
        ++cnt;
        if (num != 0 && cnt > num)
            break;
        if ((*str >= 0xD800) && (*str < 0xDC00)) {
            if ((*(str + 1) >= 0xDC00) && (*(str + 1) < 0xE000)) {
                ++cnt;
                if (num != 0 && cnt > num)
                    break;
                str += 2;
                ++ucs4_len;
            } else {
                ++str;
            }
        } else {
            ++str;
            ++ucs4_len;
        }
    }
    return ucs4_len;
}

unsigned int us_utf16_len(const utf16char *str, unsigned int max)
{
    unsigned int len = 0;
    while (*str != '\0') {
        ++str;
        ++len;
        if (max != 0 && len > max)
            return max;
    }
    return len;
}

unsigned int us_ucs4_len_of_utf8(const utf8char *str, unsigned int num)
{
    unsigned int cnt, ucs4_len;
    cnt = ucs4_len = 0;

    while (*str != '\0') {
        ++cnt;
        if (num != 0 && cnt > num)
            break;
        if ((*str & 0x80) == 0) {
            ++str;
            ++ucs4_len;
        } else if ((*str & 0xE0) == 0xC0) {
            if ((*(str + 1) & 0xC0) == 0x80) {
                ++cnt;
                if (num != 0 && cnt > num)
                    break;
                str += 2;
                ++ucs4_len;
            } else {
                ++str;
            }
        } else if ((*str & 0xF0) == 0xE0) {
            if (((*(str + 1) & 0xC0) == 0x80) && ((*(str + 2) & 0xC0) == 0x80)) {
                cnt += 2;
                if (num != 0 && cnt > num)
                    break;
                str += 3;
                ++ucs4_len;
            } else {
                ++str;
            }
        } else if ((*str & 0xF8) == 0xF0) {
            if (((*(str + 1) & 0xC0) == 0x80) && ((*(str + 2) & 0xC0) == 0x80) && ((*(str + 3) & 0xC0) == 0x80)) {
                cnt += 3;
                if (num != 0 && cnt > num)
                    break;
                str += 4;
                ++ucs4_len;
            } else {
                ++str;
            }
        } else {
            ++str;
        }
    }
    return ucs4_len;
}

void del_chr_in_str(char *str, char chr)
{
    char *ptr = str;
    while (*ptr != '\0') {
        if (*ptr != chr) {
            *str = *ptr;
            str++;
            ptr++;
        } else {
            ptr++;
        }
    }
    *str = '\0';
}

/* num은 str의 배열요소 개수이다.  널을 세지 않은 개수라도 괜찮다.
   널을 포함하지 않은 문자열을 받아도 널로 끝나는 문자열을 반환한다.  */
char *strdup_from_utf16(const utf16char *str, unsigned int num)
{
    char *ret;
    size_t ret_num;
    utf16char *temp = NULL;

    if (num == 0) {
        num = us_utf16_len(str, 0) + 1;
    } else {
        size_t temp_size = sizeof(utf16char) * (num + 1);
        temp = malloc(temp_size);
        if (temp == NULL)
            return NULL;
        memcpy(temp, str, sizeof(utf16char) * num);
        temp[num] = '\0';
        str = temp;
        ++num;
    }

    ret_num = num * 4;
    ret = malloc(ret_num);
    if (ret != NULL)
        ul_utf16_to_locale(str, num, ret, ret_num);

    free(temp);
    return ret;
}
